package me.hammerle.snuviengine.game;

import java.util.ConcurrentModificationException;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Consumer;

public class BoxList
{
    private static final double MAX_SIZE = 2.5;
    
    private final Entity[] entities = new Entity[70000];
    private int lastFreeEntity = entities.length - 1;
    public boolean isIterating = false;
    
    public static class Node
    {
        private Node next = null;
        private Node previous = null;
        private int index = -1;
        private int x = -1;
        private int y = -1;
    }
    
    private final double boxMinX;
    private final double boxMinY;
    private final double boxMaxX;
    private final double boxMaxY;
    
    private final double size;
    
    private final Node[][] nodes;
    
    private final int partionX;
    private final int partionY;
    
    public int amount = 0;
        
    public BoxList(double x1, double y1, double x2, double y2, double size)
    {
        this.size = size;
        
        boxMinX = Math.min(x1, x2);
        boxMinY = Math.min(y1, y2);
        
        partionX = (int) Math.ceil((Math.max(x1, x2) - boxMinX) / size);
        partionY = (int) Math.ceil((Math.max(y1, y2) - boxMinX) / size);
        
        boxMaxX = partionX * size;
        boxMaxY = partionY * size;
        
        System.out.println(boxMinX + " " + boxMaxX);
        System.out.println(boxMinY + " " + boxMaxY);
        System.out.println(partionX + " " + partionY);
        
        nodes = new Node[partionX][partionY];
    }
    
    public Node add(Entity ent)
    {
        if(isIterating)
        {
            throw new ConcurrentModificationException();
        }
        
        amount++;
        
        int x = Math.max(Math.min((int) (((ent.xPos + ent.width / 2) + boxMinX) / size), partionX - 1), 0);
        int y = Math.max(Math.min((int) (((ent.yPos + ent.height / 2) + boxMinY) / size), partionY - 1), 0);
        
        Node n = new Node();
        n.index = lastFreeEntity;
        n.x = x;
        n.y = y;
        
        if(nodes[x][y] == null)
        {
            nodes[x][y] = n;
        }
        else
        {
            n.next = nodes[x][y];
            nodes[x][y].previous = n;
            nodes[x][y] = n;
        }
        
        entities[lastFreeEntity--] = ent;
        
        return n;
    }
    
    public void remove(Node n)
    {
        if(isIterating)
        {
            throw new ConcurrentModificationException();
        }
        
        if(n.previous != null)
        {
            n.previous.next = n.next;
        }
        else
        {
            nodes[n.x][n.y] = n.next;
        }
        
        if(n.next != null)
        {
            n.next.previous = n.previous;
        }
        
        if(lastFreeEntity + 1 == entities.length)
        {
            entities[n.index] = null;
        }
        else
        {
            lastFreeEntity++;
            Entity ent = entities[lastFreeEntity];
            entities[n.index] = ent;
            ent.node.index = n.index;
            entities[lastFreeEntity] = null;
        }
        
        n.next = null;
        n.previous = null;
        
        amount--;
    }
    
    public void update(Entity ent)
    {
        int x = Math.max(Math.min((int) (((ent.xPos + ent.width / 2) + boxMinX) / size), partionX - 1), 0);
        int y = Math.max(Math.min((int) (((ent.yPos + ent.height / 2) + boxMinY) / size), partionY - 1), 0);
        
        Node n = ent.node;
        if(n.x == x && n.y == y)
        {
            return;
        }
        
        if(n.previous != null)
        {
            n.previous.next = n.next;
        }
        else
        {
            nodes[n.x][n.y] = n.next;
        }
        
        if(n.next != null)
        {
            n.next.previous = n.previous;
        }
        
        n.x = x;
        n.y = y;
        n.previous = null;
        
        if(nodes[x][y] == null)
        {
            nodes[x][y] = n;
            n.next = null;
        }
        else
        {
            n.next = nodes[x][y];
            nodes[x][y].previous = n;
            nodes[x][y] = n;
        }
    }
    
    public List<Entity> getAllEntitiesAt(Entity not, double minX, double minY, double maxX, double maxY)
    {
        int sx = Math.max(Math.min((int) ((minX + boxMinX - MAX_SIZE) / size), partionX - 1), 0);
        int sy = Math.max(Math.min((int) ((minY + boxMinY - MAX_SIZE) / size), partionY - 1), 0);
        int ex = Math.max(Math.min((int) ((maxX + boxMinX + MAX_SIZE) / size), partionX - 1), 0);
        int ey = Math.max(Math.min((int) ((maxY + boxMinY + MAX_SIZE) / size), partionY - 1), 0);
        
        List<Entity> list = new LinkedList<>();
        
        for(int x = sx; x <= ex; x++)
        {
            for(int y = sy; y <= ey; y++)
            {
                Node n = nodes[x][y];
                while(n != null)
                {
                    if(not != entities[n.index] && entities[n.index].isColliding(minX, minY, maxX, maxY))
                    {
                        list.add(entities[n.index]);
                    }
                    n = n.next;
                }
            }
        }
        
        return list;
    } 
    
    public void forEach(Consumer<Entity> c)
    {
        isIterating = true;
        for(int i = lastFreeEntity + 1; i < entities.length; i++)
        {
            c.accept(entities[i]);
        }
        isIterating = false;
    }

    @Override
    public String toString()
    {
        StringBuilder sb = new StringBuilder();
        for(Entity ent : entities)
        {
            sb.append(ent);
            sb.append(", ");
        }
        return sb.toString();
    }
}
